# Features modeled after my Rust lib https://crates.io/crates/human-repr

class Features:
    def __init__(self):
        self._feature_space = False
        self._feature_1024 = False
        self._feature_iec = False

    @property
    def feature_space(self) -> bool:
        return self._feature_space

    @property
    def feature_1024(self) -> bool:
        return self._feature_1024

    @property
    def feature_iec(self) -> bool:
        return self._feature_iec

    @feature_space.setter
    def feature_space(self, value: bool):  # pragma: no cover
        self._feature_space = bool(value)

    @feature_1024.setter
    def feature_1024(self, value: bool):  # pragma: no cover
        self._feature_1024 = bool(value)

    @feature_iec.setter
    def feature_iec(self, value: bool):  # pragma: no cover
        self._feature_iec = bool(value)
        self.feature_1024 = value


def conv_space(space: bool) -> str:
    return {False: '', True: ' '}[space]


FEATURES = Features()



from typing import Optional

from .features import FEATURES, conv_space

SPEC = (
    (1e3, 1e3, "ns", 1),
    (1e3, 1e3, "µs", 1),  # uses non-ASCII “µs” suffix.
    (1e3, 1e3, "ms", 1),
    (60., 1., "s", 2),
    # 1:01.1 (minutes in code, 1 decimal).
    # 1:01:01 (hours in code, 0 decimal).
)


def __human_duration(val: float, prec: Optional[int], space: str) -> str:
    val *= 1e9
    for size, div_next, scale, dec in SPEC:
        r = round(val, dec)
        if r >= size:
            val /= div_next
            continue

        if prec is not None:
            r = round(val, prec)
        elif r % 1. == 0.:
            prec = 0
        elif (r * 10.) % 1. == 0.:
            prec = 1
        else:
            prec = 2
        return '{:.{}f}{}{}'.format(r, prec, space, scale)

    val = round(val, 1)
    m = val / 60.
    if m < 60.:
        r = val % 60.
        if prec is not None:
            pass
        elif r % 1. == 0.:
            prec = 0

        if prec == 0:
            return '{:.0f}:{:02.0f}'.format(m // 1., r)
        return '{:.0f}:{:04.1f}'.format(m // 1., round(r, 1))

    return '{:.0f}:{:02.0f}:{:02.0f}'.format(m / 60. // 1., m % 60. // 1., val % 60. // 1.)


def fn_human_duration(space: bool):
    def run(val, prec: Optional[int]):
        return __human_duration(val, prec, space)

    space = conv_space(space)
    return run


class HumanDuration(object):
    def __init__(self, value):
        assert value >= 0.
        self._value = value

    @property
    def value(self):
        return self._value

    def as_human(self, prec: Optional[int] = None) -> str:
        """Return a beautiful representation of this duration.
        It dynamically calculates the best scale to use.

        Args:
            prec: an optional custom precision

        Returns:
            the human friendly representation.

        """
        return fn_human_duration(FEATURES.feature_space)(self._value, prec)

    def __str__(self):
        return self.as_human()

    def __repr__(self):  # pragma: no cover
        return 'HumanDuration{{ value={} }} -> {}'.format(self._value, self)

    def __eq__(self, other):
        return self.__str__() == other



from typing import Optional

from .features import FEATURES, conv_space
from .human_count import fn_human_count

SPEC = (
    (24., "/d", 2),
    (60., "/h", 1),
    (60., "/m", 1),
    # "/s" in code.
)


def __human_throughput(val: float, unit: str, prec: Optional[int], space: str, fn_count) -> str:
    val *= 60. * 60. * 24.
    for size, scale, dec in SPEC:
        r = round(val, dec)
        if r >= size:
            val /= size
            continue

        if prec is not None:
            r = round(val, prec)
        elif r % 1. == 0.:
            prec = 0
        elif (r * 10.) % 1. == 0.:
            prec = 1
        else:
            prec = 2
        return '{:.{}f}{}{}{}'.format(r, prec, space, unit, scale)

    return '{}/s'.format(fn_count(val, unit, prec))


def fn_human_throughput(space: bool, d1024: bool, iec: bool):
    def run(val: float, unit: str, prec: Optional[int]):
        return __human_throughput(val, unit, prec, space, fn_count)

    fn_count = fn_human_count(space, d1024, iec)
    space = conv_space(space)
    return run


class HumanThroughput(object):
    def __init__(self, value, unit):
        assert value >= 0.
        self._value = value
        self._unit = unit

    @property
    def value(self):
        return self._value

    def unit(self, value: str) -> 'HumanThroughput':
        self._unit = value
        return self

    def as_human(self, prec: Optional[int] = None) -> str:
        """Return a beautiful representation of this count.
        It dynamically calculates the best scale to use.

        Args:
            prec: an optional custom precision

        Returns:
            the human friendly representation.

        """
        return fn_human_throughput(FEATURES.feature_space, FEATURES.feature_1024,
                                   FEATURES.feature_iec)(self._value, self._unit, prec)

    def __str__(self):
        return self.as_human()

    def __repr__(self):  # pragma: no cover
        return 'HumanCount{{ value={} }} -> {}'.format(self._value, self)

    def __eq__(self, other):
        return self.__str__() == other



import time
from contextlib import contextmanager

from .human_count import HumanCount
from .human_duration import HumanDuration
from .human_throughput import HumanThroughput


def about_time(func_or_it=None, *args, **kwargs):
    """Measure timing and throughput of code blocks, with beautiful
    human friendly representations.

    There are three modes of operation: context manager, callable and
    throughput.

    1. Use it like a context manager:

    >>> with about_time() as t:
    ....    # code block.

    2. Use it with a callable:

    >>> def func(a, b): ...
    >>> t = about_time(func, 1, b=2)  # send arguments at will.

    3. Use it with an iterable or generator:

    >>> t = about_time(it)  # any iterable or generator.
    >>> for item in t:
    ....    # use item
    """

    timings = [0.0, 0.0]

    # use as a context manager.
    if func_or_it is None:
        return _context_timing(timings, Handle(timings))

    # use as a callable.
    if callable(func_or_it):
        with _context_timing(timings):
            result = func_or_it(*args, **kwargs)
        return HandleResult(timings, result)

    try:
        it = iter(func_or_it)
    except TypeError:
        raise UserWarning('param should be callable or iterable.')

    # use as a counter/throughput iterator.
    def it_closure():
        with _context_timing(timings):
            for it_closure.count, elem in enumerate(it, 1):  # iterators are iterable.
                yield elem

    it_closure.count = 0  # the count will only be updated after starting iterating.
    return HandleStats(timings, it_closure)


@contextmanager
def _context_timing(timings, handle=None):
    timings[0] = time.perf_counter()
    yield handle
    timings[1] = time.perf_counter()


class Handle(object):
    def __init__(self, timings):
        self.__timings = timings

    @property
    def duration(self) -> float:
        """Return the actual duration in seconds.
        This is dynamically updated in real time.

        Returns:
            the number of seconds.

        """
        return (self.__timings[1] or time.perf_counter()) - self.__timings[0]

    @property
    def duration_human(self) -> HumanDuration:
        """Return a beautiful representation of the duration.
        It dynamically calculates the best unit to use.

        Returns:
            the human representation.

        """
        return HumanDuration(self.duration)


class HandleResult(Handle):
    def __init__(self, timings, result):
        super(HandleResult, self).__init__(timings)
        self.__result = result

    @property
    def result(self):
        """Return the result of the callable.

        Returns:
            the result of the callable.

        """
        return self.__result


class HandleStats(Handle):
    def __init__(self, timings, it_closure):
        super(HandleStats, self).__init__(timings)
        self.__it = it_closure

    def __iter__(self):
        return self.__it()

    @property
    def count(self) -> int:
        """Return the current iteration count.
        This is dynamically updated in real time.

        Returns:
            the current iteration count.

        """
        return self.__it.count

    @property
    def count_human(self) -> HumanCount:
        """Return a beautiful representation of the current iteration count.
        This is dynamically updated in real time.

        Returns:
            the human representation.

        """
        return self.count_human_as('')

    def count_human_as(self, unit: str) -> HumanCount:
        """Return a beautiful representation of the current iteration count.
        This is dynamically updated in real time.

        Args:
            unit: what is being measured

        Returns:
            the human representation.

        """
        return HumanCount(self.count, unit)

    @property
    def throughput(self) -> float:
        """Return the current throughput in items per second.
        This is dynamically updated in real time.

        Returns:
            the number of items per second.

        """
        try:
            return self.count / self.duration
        except ZeroDivisionError:  # pragma: no cover
            return float('nan')

    @property
    def throughput_human(self) -> HumanThroughput:
        """Return a beautiful representation of the current throughput.
        It dynamically calculates the best unit to use.

        Returns:
            the human representation.

        """
        return self.throughput_human_as('')

    def throughput_human_as(self, unit: str) -> HumanThroughput:
        """Return a beautiful representation of the current throughput.
        It dynamically calculates the best unit to use.

        Args:
            unit: what is being measured

        Returns:
            the human representation.

        """
        return HumanThroughput(self.throughput, unit)



from .core import about_time
from .features import FEATURES
from .human_count import HumanCount
from .human_duration import HumanDuration
from .human_throughput import HumanThroughput

VERSION = (4, 2, 1)

__author__ = 'Rogério Sampaio de Almeida'
__email__ = 'rsalmei@gmail.com'
__version__ = '.'.join(map(str, VERSION))

__all__ = ('__author__', '__version__', 'about_time', 'HumanCount', 'HumanDuration',
           'HumanThroughput', 'FEATURES')



from typing import Optional

from .features import FEATURES, conv_space

SI_1000_SPEC = ('', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
SI_1024_SPEC = ('', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
IEC_1024_SPEC = ('', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi')
DECIMALS = [1, 1, 1, 2, 2, 2, 2, 2, 2]


def __human_count(val: float, unit: str, prec: Optional[int], space: str, divisor: int,
                  spec: tuple) -> str:
    for scale, dec in zip(spec, DECIMALS):
        r = round(val, dec)
        if r >= divisor:
            val /= divisor
            continue
        break
    else:
        r, scale = val, '+'

    if prec is not None:
        r = round(val, prec)
    elif r % 1. == 0.:
        prec = 0
    elif (r * 10.) % 1. == 0.:
        prec = 1
    else:
        prec = 2
    return '{:.{}f}{}{}{}'.format(r, prec, space, scale, unit)


def fn_human_count(space: bool, d1024: bool, iec: bool):
    def run(val: float, unit: str, prec: Optional[int]):
        return __human_count(val, unit, prec, space, divisor, spec)

    space = conv_space(space)
    divisor, spec = {
        (False, False): (1000, SI_1000_SPEC),
        (True, False): (1024, SI_1024_SPEC),
        (True, True): (1024, IEC_1024_SPEC),
        (False, True): (1024, IEC_1024_SPEC),  # invalid combination, which just returns the above.
    }[(d1024, iec)]
    return run


class HumanCount(object):
    def __init__(self, value, unit):
        assert value >= 0.
        self._value = value
        self._unit = unit

    @property
    def value(self):
        return self._value

    def unit(self, value: str) -> 'HumanCount':
        self._unit = value
        return self

    def as_human(self, prec: Optional[int] = None) -> str:
        """Return a beautiful representation of this count.
        It dynamically calculates the best scale to use.

        Args:
            prec: an optional custom precision

        Returns:
            the human friendly representation.

        """
        return fn_human_count(FEATURES.feature_space, FEATURES.feature_1024,
                              FEATURES.feature_iec)(self._value, self._unit, prec)

    def __str__(self):
        return self.as_human()

    def __repr__(self):  # pragma: no cover
        return 'HumanCount{{ value={} }} -> {}'.format(self._value, self)

    def __eq__(self, other):
        return self.__str__() == other
